#include "RPi_Pico_TimerInterrupt.h"
#include "hardware/pwm.h"
#include <LittleFS.h>

//44k 22->17
#define TIMER1_INTERVAL_us 17
//32k 31 -> 26
//#define TIMER1_INTERVAL_us 26
//16k 62 -> 57
//#define TIMER1_INTERVAL_us 57
//8k 125 -> 120
//#define TIMER1_INTERVAL_us 120

#define PIN_BUZZER    A0
#define PIN_DEBUG_PWM    D12
#define PIN_DEBUG_MANAGE    D11

//bufferの大きさ
//#define BUF_LENGTH 32768
//#define BUF_LENGTH 16384
//#define BUF_LENGTH 8192
#define BUF_LENGTH 4096
//#define BUF_LENGTH 2048
//#define BUF_LENGTH 1024

//和音数
#define WAVE_CHORDS 6

//waveのオフセット
#define WAVE_OFFSET 44

//wave_modeのenum
#define WAVE_MODE_STOP 0
#define WAVE_MODE_RUN 1
#define WAVE_MODE_NONE 2


#define BUF_CH_A 0
#define BUF_CH_B 1

// Init RPI_PICO_Timer
RPI_PICO_Timer ITimer1(1);

//各chの音
uint32_t wave_position[WAVE_CHORDS];//現在値
uint32_t wave_length[WAVE_CHORDS];//長さ
uint16_t wave_volume[WAVE_CHORDS];//volume 0-255
String wave_name[WAVE_CHORDS];//ファイル名
uint8_t wave_last_val[WAVE_CHORDS];//プチ音を消すため最後の音を保つ。
int wave_status[WAVE_CHORDS];//状態　WAVE_MODE_STOP,WAVE_MODE_RUN
int wave_order[WAVE_CHORDS];//要求　WAVE_MODE_STOP,WAVE_MODE_RUN

//出力部分
volatile uint slice_num = 0;
volatile int16_t wave_buf_a[BUF_LENGTH];
volatile int16_t wave_buf_b[BUF_LENGTH];
volatile int buf_ab = 1;
volatile uint16_t buf_count = 0;
volatile bool buf_flag;//false:次のバッファを書かないといけない、true:書き込み済み

File f;

String inputString = "";         // a String to hold incoming data
bool stringComplete = false;  // whether the string is complete

bool TimerHandler1(struct repeating_timer *t) {
  //割込みで呼ばれるところ
  digitalWrite(PIN_DEBUG_PWM,HIGH);
  buf_count ++;
  //bufferの切り替え確認
  if(buf_count == BUF_LENGTH){
    if(buf_ab == BUF_CH_A){
      buf_ab = BUF_CH_B;
    }
    else{
      buf_ab = BUF_CH_A;
    }
    buf_flag  = false;
    buf_count  = 0;
  }
  //PWMへの書き込み A,Bで分かれる。
  if(buf_ab == BUF_CH_A){
    pwm_set_chan_level(slice_num, PWM_CHAN_A, wave_buf_a[buf_count]);
  }
  else{
    pwm_set_chan_level(slice_num, PWM_CHAN_A, wave_buf_b[buf_count]);
  }
  /*
  if((buf_count > BUF_LENGTH -10) || (buf_count < 50))
  {
    Serial.print(buf_ab);
    Serial.print(",");
    Serial.print(buf_count);
    Serial.print(",");
    Serial.print(wave_buf_a[buf_count]);
    Serial.print(",");
    Serial.println(wave_buf_b[buf_count]);
  }
  */
  digitalWrite(PIN_DEBUG_PWM,LOW);
  return true;
}

void manage_wave()
{
  //バッファへの書き込みなど
  
  //書く必要があるか？
  if(buf_flag == true){
    return;
  }

  digitalWrite(PIN_DEBUG_MANAGE,HIGH);

  //当該バッファを０リセット
  if(buf_ab == BUF_CH_A){
    for(int i = 0; i < BUF_LENGTH;i++){
      wave_buf_b[i] = 0;
    }
  }
  else{
    for(int i = 0; i < BUF_LENGTH;i++){
      wave_buf_a[i] = 0;
    }
  }
  
  //buffer
  uint8_t temp_buf[BUF_LENGTH];

  for(int i = 0;i < WAVE_CHORDS;i++){
    //buffer初期化
    for(int j = 0; j < BUF_LENGTH;j++)
    {
      temp_buf[j] = wave_last_val[i];
      //temp_buf[i] = 0;
    }
    //orderによりやることを変える
    switch(wave_order[i])
    {
      case WAVE_MODE_RUN:
        //鳴らすとき bufferからロード
        wave_order[i] = WAVE_MODE_NONE;
        
        f = LittleFS.open(wave_name[i],"r");      
        wave_length[i] = f.size() - WAVE_OFFSET;
        wave_position[i] = 0;
        f.seek(wave_position[i] + WAVE_OFFSET,SeekSet);
        if((wave_length[i] - wave_position[i]) < BUF_LENGTH){
          f.read(temp_buf,(wave_length[i] - wave_position[i]));
          wave_last_val[i] = temp_buf[wave_length[i] - wave_position[i] - 2];
          //Serial.print("first:");
          //Serial.println(wave_last_val[i]);
          //残りを同じ値で埋めておく。
          for(int j = wave_length[i] - wave_position[i] - 1;j < BUF_LENGTH;j++){
            temp_buf[j] = wave_last_val[i];
          }
          wave_status[i] = WAVE_MODE_STOP;
        }
        else{
          f.read(temp_buf,BUF_LENGTH);
          wave_last_val[i] = temp_buf[BUF_LENGTH-1];
          //Serial.print("first:");
          //Serial.println(wave_last_val[i]);
          wave_status[i] = WAVE_MODE_RUN;
        }
        wave_position[i] = f.position() - WAVE_OFFSET;
        
        //Serial.println(wave_position[i]);
        f.close();
        
        break;
      case WAVE_MODE_STOP:
        //次のターンで消す
        wave_order[i] = WAVE_MODE_NONE;
        wave_status[i] = WAVE_MODE_STOP;
        wave_position[i] = 0;
        wave_length[i] = 0;
        wave_volume[i] = 255;
        //特になにもせずに無音になる。127を入れているので
        break;
      case WAVE_MODE_NONE:
        //作業を続ける
        wave_order[i] = WAVE_MODE_NONE;
        if(wave_status[i] == WAVE_MODE_STOP){
          break;
        }
          
        f = LittleFS.open(wave_name[i],"r");      
        //wave_length[i] = f.size() - WAVE_OFFSET;
        //wave_position[i] = 0;
        f.seek(wave_position[i] + WAVE_OFFSET,SeekSet);
        if((wave_length[i] - wave_position[i]) < BUF_LENGTH){
          f.read(temp_buf,(wave_length[i] - wave_position[i]));
          wave_last_val[i] = temp_buf[wave_length[i] - wave_position[i] - 2];
          //Serial.print("last:");
          //Serial.println(wave_length[i] - wave_position[i]);
          //残りを同じ値で埋めておく。
          for(int j = wave_length[i] - wave_position[i] - 1;j < BUF_LENGTH;j++){
            temp_buf[j] = wave_last_val[i];
          }
          /*
          for (int k = wave_length[i] - wave_position[i] - 50;k < wave_length[i] - wave_position[i] + 50;k++)
          {
            Serial.print(k);
            Serial.print(",");
            Serial.println(temp_buf[k]);
          }
          */
          wave_status[i] = WAVE_MODE_STOP;
        }
        else{
          f.read(temp_buf,BUF_LENGTH);
          wave_last_val[i] = temp_buf[BUF_LENGTH-1];
          //Serial.print("continue:");
          //Serial.println(wave_last_val[i]);
          wave_status[i] = WAVE_MODE_RUN;
        }
        wave_position[i] = f.position() - WAVE_OFFSET;
        //Serial.println(wave_position[i]);
        f.close();
        
        break;
      default:
        break;
    }
    
    //音量を変更しマスターバッファに入れる(音量変更はそのうち）
    if(buf_ab == BUF_CH_A){
      for(int i = 0; i < BUF_LENGTH;i++){
        //wave_buf_b[i] += (uint16_t)temp_buf[i] * wave_volume[i] / 255;
        wave_buf_b[i] += temp_buf[i];
      }
    }
    else{
      for(int i = 0; i < BUF_LENGTH;i++){
        //wave_buf_a[i] += (uint16_t)temp_buf[i] * wave_volume[i] / 255;
        wave_buf_a[i] += temp_buf[i];
      }
    }
    
  }
  buf_flag = true;
  digitalWrite(PIN_DEBUG_MANAGE,LOW);
  return;
}

void play_wav(int8_t ch,int16_t freq) {
  wave_order[ch] = WAVE_MODE_RUN;
}
void play_cls() {
  for(int i = 0;i < WAVE_CHORDS;i++){
    wave_order[i] = WAVE_MODE_STOP;
  }
}

void serialEvent() {
  while (Serial.available()) {
    // get the new byte:
    char inChar = (char)Serial.read();
    if (inChar == '\n') {
      stringComplete = true;
      return;
    }
    // add it to the inputString:
    inputString += inChar;
  }
}

void setup() {
  // put your setup code here, to run once:
  ITimer1.attachInterruptInterval(TIMER1_INTERVAL_us, TimerHandler1);
  gpio_set_function(PIN_BUZZER, GPIO_FUNC_PWM);
  slice_num = pwm_gpio_to_slice_num(PIN_BUZZER);
  pwm_set_wrap(slice_num, 1023);
  pwm_set_chan_level(slice_num, PWM_CHAN_A, 0);
  //pwm_set_chan_level(slice_num, PWM_CHAN_B, 0);

  //debug pin 
  pinMode(PIN_DEBUG_PWM, OUTPUT);
  pinMode(PIN_DEBUG_MANAGE, OUTPUT);
  digitalWrite(PIN_DEBUG_PWM,LOW);
  digitalWrite(PIN_DEBUG_MANAGE,LOW);

  LittleFS.begin();

  //初期化
  for(int i = 0;i < WAVE_CHORDS;i++) {
    wave_position[i] = 0;
    wave_length[i] = 0;
    wave_volume[i] = 255;
    wave_last_val[i] = 127;
    wave_status[i] = WAVE_MODE_STOP;
    wave_order[i] = WAVE_MODE_NONE;
  }
  //String dir = "/8kHz_8/";
  //String dir = "/16kHz_8/";
  //String dir = "/32kHz_8/";
  String dir = "/";
  wave_name[0] = dir + "185_CP_C2000.wav";
  wave_name[1] = dir + "185_door_open2.wav";
  wave_name[2] = dir + "185_shanai_chime.wav";
  wave_name[3] = dir + "185_shinagawa_B.wav";
  wave_name[4] = dir + "CrossingCountry.wav";
  wave_name[5] = dir + "Station_Announce01.wav";

  //当該バッファを０リセット
  for(int i = 0; i < BUF_LENGTH;i++){
    wave_buf_a[i] = 0;
    wave_buf_b[i] = 0;
  }
  Serial.begin(115200);
  inputString.reserve(200);
        
  pwm_set_enabled(slice_num, true);

}

void loop() {
  // put your main code here, to run repeatedly:
  if (stringComplete) {
    Serial.println(inputString);
    if(inputString.equals("0")){play_wav(0,16);}
    else if(inputString.equals("1")){play_wav(1,16);}
    else if(inputString.equals("2")){play_wav(2,16);}
    else if(inputString.equals("3")){play_wav(3,16);}
    else if(inputString.equals("4")){play_wav(4,16);}
    else if(inputString.equals("5")){play_wav(5,16);}
    else if(inputString.equals("c")){play_cls();}
    // clear the string:
    inputString = "";
    stringComplete = false;
  }
  manage_wave();
  /*
  Serial.print(buf_ab);
  Serial.print(",");
  Serial.println(buf_count);
  */
}
